@@ -20,16 +20,6 @@
 
 open Js
 
-class type tokenList = object
-  method length : int readonly_prop
-  method item : int -> js_string t optdef meth
-  method contains : js_string t -> bool t meth
-  method add : js_string t -> unit meth
-  method remove : js_string t -> unit meth
-  method toggle : js_string t -> bool t meth
-  method stringifier : js_string t prop
-end
-
 class type ['node] nodeList = object
   method item : int -> 'node t opt meth
   method length : int readonly_prop
@@ -62,6 +52,13 @@ type nodeType =
   | DOCUMENT_FRAGMENT
   | NOTATION
 
+let document_position_disconnected = 0x01
+let document_position_preceding    = 0x02
+let document_position_following    = 0x04
+let document_position_contains     = 0x08
+let document_position_contained_by = 0x10
+let document_position_implementation_specific = 0x20
+
 class type node = object
   method nodeName : js_string t readonly_prop
   method nodeValue : js_string t opt readonly_prop
@@ -79,6 +76,7 @@ class type node = object
   method appendChild : node t -> node t meth
   method hasChildNodes : bool t meth
   method cloneNode : bool t -> node t meth
+  method compareDocumentPosition : node t -> int meth
 end
 
 let appendChild (p : #node t) (n : #node t) =
@@ -103,11 +101,11 @@ class type attr = object
 end
 
 (** Specification of [NamedNodeMap] objects. *)
-and namedNodeMap = object
-  method getNamedItem : js_string t -> node t opt meth
-  method setNamedItem : node t -> node t opt meth
-  method removeNamedItem : js_string t -> node t opt meth
-  method item : int -> node t opt meth
+and ['node] namedNodeMap = object
+  method getNamedItem : js_string t -> 'node t opt meth
+  method setNamedItem : 'node t -> 'node t opt meth
+  method removeNamedItem : js_string t -> 'node t opt meth
+  method item : int -> 'node t opt meth
   method length : int readonly_prop
 end
 
@@ -120,8 +118,7 @@ and element = object
   method removeAttribute : js_string t -> unit meth
   method hasAttribute : js_string t -> bool t meth
   method getElementsByTagName : js_string t -> element nodeList t meth
-  method classList : tokenList t readonly_prop
-  method attributes : namedNodeMap t readonly_prop
+  method attributes : attr namedNodeMap t readonly_prop
 end
 
 class type characterData = object
@@ -132,7 +129,7 @@ class type characterData = object
   method appendData : js_string t -> unit meth
   method insertData : int -> js_string t -> unit meth
   method deleteData : int -> int -> unit meth
-  method replaceData : int -> int -> js_string t meth
+  method replaceData : int -> int -> js_string t -> unit meth
 end
 
 class type text = characterData
@@ -183,3 +180,93 @@ module CoerceTo = struct
   let attr e : attr Js.t Js.opt = cast e ATTRIBUTE
 
 end
+
+type ('a, 'b) event_listener = ('a, 'b -> bool t) meth_callback opt
+  (** The type of event listener functions.  The first type parameter
+      ['a] is the type of the target object; the second parameter
+      ['b] is the type of the event object. *)
+
+class type ['a] event = object
+  method _type : js_string t readonly_prop
+  method target : 'a t optdef readonly_prop
+  method currentTarget : 'a t optdef readonly_prop
+
+  (* Legacy methods *)
+  method srcElement : 'a t optdef readonly_prop
+end
+
+let no_handler : ('a, 'b) event_listener = Js.null
+let window_event () : 'a #event t = Js.Unsafe.variable "event"
+(* The function preventDefault must be called explicitely when
+   using addEventListener... *)
+let handler f =
+  Js.some (Js.wrap_callback
+    (fun e ->
+      (* depending on the internet explorer version, e can be 0, null
+	 or undefined. This is the only way I know to test them all *)
+      if not (Obj.magic e)
+      then
+        let e = window_event () in
+        let res = f e in
+        e##returnValue <- res;
+	res
+      else
+	let res = f e in
+        if not (Js.to_bool res) then
+          (Js.Unsafe.coerce e)##preventDefault ();
+        res))
+let full_handler f =
+  Js.some (Js.wrap_meth_callback
+    (fun this e ->
+      (* depending on the internet explorer version, e can be 0, null
+	 or undefined. This is the only way I know to test them all *)
+      if not (Obj.magic e)
+      then
+        let e = window_event () in
+        let res = f this e in
+        e##returnValue <- res; res
+      else
+        let res = f this e in
+        if not (Js.to_bool res) then
+          (Js.Unsafe.coerce e)##preventDefault ();
+        res))
+let invoke_handler
+  (f : ('a, 'b) event_listener) (this : 'a) (event : 'b) : bool t =
+  Js.Unsafe.call f this [|Js.Unsafe.inject event|]
+
+let node_constr : node t constr = Js.Unsafe.variable "window.Node"
+
+let eventTarget (e: (< .. > as 'a) #event t) : 'a t =
+  let target =
+    Optdef.get (e##target) (fun () ->
+    Optdef.get (e##srcElement) (fun () -> assert false))
+  in
+  if Js.instanceof target node_constr
+  then
+    (* Workaround for Safari bug *)
+    let target' : node Js.t = Js.Unsafe.coerce target in
+    if target'##nodeType == TEXT then
+      Js.Unsafe.coerce (Opt.get (target'##parentNode) (fun () -> assert false))
+    else
+      target
+  else target
+
+module Event = struct
+  type 'a typ = Js.js_string Js.t
+  let make s = Js.string s
+end
+
+type event_listener_id = unit -> unit
+
+let addEventListener (e : (< .. > as 'a) t) typ h capt =
+  if (Js.Unsafe.coerce e)##addEventListener == Js.undefined then begin
+    let ev = (Js.string "on")##concat(typ) in
+    let callback = fun e -> Js.Unsafe.call (h, e, [||]) in
+    (Js.Unsafe.coerce e)##attachEvent(ev, callback);
+    fun () -> (Js.Unsafe.coerce e)##detachEvent(ev, callback)
+  end else begin
+    (Js.Unsafe.coerce e)##addEventListener(typ, h, capt);
+    fun () -> (Js.Unsafe.coerce e)##removeEventListener (typ, h, capt)
+  end
+
+let removeEventListener id = id ()
